DETECTION_TYPE = "optical-imaging"
SOFTWARE_NAME = "Apex"
DEVICE_TYPE = "imager"

IMAGE_FEATURES = [
    "Spot Forming Units (SFU)",
    "Sum of Spot Volume (RSV)",
    "Average Relative Spot Volume (RSV)",
    "Average Relative Spot Volume {filter} (RSV{led_number})",
    "Sum Relative Spot Volume {filter} (RSV{led_number})",
]
