from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DISPLAY_NAME = "Bio-Rad CFX Maestro"
DEVICE_TYPE = "qPCR"

PRODUCT_MANUFACTURER = "Bio-Rad"
SOFTWARE_NAME = "CFX Maestro"
CONTAINER_TYPE = ContainerType.well_plate
