DEVICE_IDENTIFIER = "Nanodrop"
DEVICE_TYPE = "absorbance detector"
MODEL_NUBMER = "Nanodrop 8000"

# These may be reported in the results, and are stored as calculated data
ABSORBANCE_RATIOS = [
    (260, 280),
    (260, 230),
]
