from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DEVICE_TYPE = "qPCR"
PRODUCT_MANUFACTURER = "ThermoFisher Scientific"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.qPCR_reaction_block
