from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DEVICE_TYPE = "qPCR"
SOFTWARE_NAME = "Thermo QuantStudio"
SOFTWARE_VERSION = "1.0"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.qPCR_reaction_block

POSSIBLE_WELL_COUNTS = [1, 2, 4, 6, 8, 12, 24, 48, 72, 96, 384, 1536, 3456]
